from __future__ import annotations


from .semantics import Modifier, SemanticStructure


ALL = "*"


def step_path_indexer(value) -> str:
    try:
        return (
            f"Step:{''.join(name.value for key, name in value if key.value == 'label')}"
        )
    except Exception as err:
        raise Exception(f"Failed to process '{value}', with {err}")


def function_ref_indexer(value) -> tuple[str, str] | None:
    try:
        kind = None
        name = None
        for key, key_value in value:
            if key.value == "kind":
                kind = key_value.value

            if key.value == "name":
                name = key_value.value

        if not (kind and name):
            return None

        return ("name", f"{kind}:{name}:ref")

    except Exception as err:
        raise Exception(f"Failed to process '{value}', with {err}")


_api_version: SemanticStructure = SemanticStructure(
    sub_structure=SemanticStructure(
        local_key_fn=lambda value: f"api_version",
        type="namespace",
    )
)

_kind: SemanticStructure = SemanticStructure(
    sub_structure=SemanticStructure(
        local_key_fn=lambda value: f"kind",
        type="type",
    ),
)

_namespace: SemanticStructure = SemanticStructure(
    sub_structure=SemanticStructure(
        type="namespace",
    ),
)

_logic_ref: SemanticStructure = SemanticStructure(
    type="property",
    sub_structure=SemanticStructure(
        strict_sub_structure_keys=True,
        field_index_key_fn=function_ref_indexer,
        sub_structure={
            "kind": SemanticStructure(
                type="property",
                sub_structure=SemanticStructure(type="type"),
            ),
            "name": SemanticStructure(
                type="property",
                sub_structure=SemanticStructure(
                    type="function",
                ),
            ),
        },
    ),
)

_api_config: SemanticStructure = SemanticStructure(
    type="property",
    strict_sub_structure_keys=True,
    sub_structure={
        "apiVersion": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="namespace"),
        ),
        "kind": SemanticStructure(
            type="parameter", sub_structure=SemanticStructure(type="type")
        ),
        "plural": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(
                type="number",
            ),
        ),
        "namespaced": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="number"),
        ),
        "owned": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="number"),
        ),
        "deleteIfExists": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="number"),
        ),
        "readonly": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="number"),
        ),
        "name": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="string"),
        ),
        "namespace": SemanticStructure(
            type="parameter",
            sub_structure=SemanticStructure(type="string"),
        ),
    },
)

_function_inputs: SemanticStructure = SemanticStructure(
    type="property",
    local_key_fn=lambda value: "inputs",
    sub_structure=SemanticStructure(
        local_key_fn=lambda value: "input-values",
        sub_structure={
            ALL: SemanticStructure(
                local_key_fn=lambda value: f"input:{value}", type="variable"
            )
        },
    ),
)

_function_input_overrides: SemanticStructure = SemanticStructure(
    type="property",
    local_key_fn=lambda value: "input_overrides",
    sub_structure=SemanticStructure(
        local_key_fn=lambda value: "input-values",
        sub_structure={
            ALL: SemanticStructure(
                local_key_fn=lambda value: f"input:{value}", type="variable"
            )
        },
    ),
)


_validators = SemanticStructure(
    type="property",
    strict_sub_structure_keys=True,
    sub_structure={
        "assert": SemanticStructure(
            type="property",
            sub_structure=SemanticStructure(type="string"),
        ),
        "ok": SemanticStructure(
            type="property", strict_sub_structure_keys=True, sub_structure={}
        ),
        "skip": SemanticStructure(
            type="property",
            strict_sub_structure_keys=True,
            sub_structure={
                "message": SemanticStructure(
                    type="property",
                    sub_structure=SemanticStructure(type="string"),
                ),
            },
        ),
        "depSkip": SemanticStructure(
            type="property",
            strict_sub_structure_keys=True,
            sub_structure={
                "message": SemanticStructure(
                    type="property",
                    sub_structure=SemanticStructure(type="string"),
                ),
            },
        ),
        "retry": SemanticStructure(
            type="property",
            strict_sub_structure_keys=True,
            sub_structure={
                "message": SemanticStructure(
                    type="property",
                    sub_structure=SemanticStructure(type="string"),
                ),
                "delay": SemanticStructure(
                    type="property",
                    sub_structure=SemanticStructure(type="number"),
                ),
            },
        ),
        "permFail": SemanticStructure(
            type="property",
            strict_sub_structure_keys=True,
            sub_structure={
                "message": SemanticStructure(
                    type="property",
                    sub_structure=SemanticStructure(type="string"),
                ),
            },
        ),
    },
)

SEMANTIC_TYPE_STRUCTURE: dict[str, SemanticStructure] = {
    "ResourceFunction": SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure={
                    "name": SemanticStructure(
                        sub_structure=SemanticStructure(
                            index_key_fn=lambda value: f"ResourceFunction:{value}:def",
                            type="function",
                            modifier=[Modifier.definition],
                        ),
                    ),
                    "namespace": _namespace,
                },
            ),
            "spec": SemanticStructure(
                strict_sub_structure_keys=True,
                sub_structure={
                    "preconditions": _validators,
                    "locals": SemanticStructure(
                        type="property",
                        sub_structure=SemanticStructure(
                            sub_structure=SemanticStructure(
                                type="property",
                            ),
                        ),
                    ),
                    "apiConfig": _api_config,
                    "resourceTemplateRef": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "name": SemanticStructure(
                                type="property",
                                sub_structure=SemanticStructure(
                                    type="string",
                                    index_key_fn=lambda value: (
                                        None
                                        if value.startswith("=")
                                        else f"ResourceTemplate:{value}:ref"
                                    ),
                                ),
                            ),
                        },
                    ),
                    "resource": SemanticStructure(
                        type="property",
                    ),
                    "overlays": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure=SemanticStructure(
                            type="property",
                            strict_sub_structure_keys=True,
                            sub_structure={
                                "skipIf": SemanticStructure(
                                    type="property",
                                ),
                                "overlay": SemanticStructure(
                                    type="property",
                                ),
                                "overlayRef": _logic_ref,
                                "inputs": _function_inputs,
                            },
                        ),
                    ),
                    "create": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "enabled": SemanticStructure(
                                type="parameter",
                                sub_structure=SemanticStructure(type="number"),
                            ),
                            "delay": SemanticStructure(
                                type="parameter",
                                sub_structure=SemanticStructure(type="number"),
                            ),
                            "overlay": SemanticStructure(
                                type="property",
                            ),
                        },
                    ),
                    "update": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "patch": SemanticStructure(
                                type="parameter",
                                strict_sub_structure_keys=True,
                                sub_structure={
                                    "delay": SemanticStructure(type="number")
                                },
                            ),
                            "recreate": SemanticStructure(
                                type="parameter",
                                strict_sub_structure_keys=True,
                                sub_structure={
                                    "delay": SemanticStructure(type="number")
                                },
                            ),
                            "never": SemanticStructure(
                                type="parameter",
                                strict_sub_structure_keys=True,
                                sub_structure={},
                            ),
                        },
                    ),
                    "delete": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "abandon": SemanticStructure(
                                type="parameter",
                                strict_sub_structure_keys=True,
                                sub_structure={},
                            ),
                            "destroy": SemanticStructure(
                                type="parameter",
                                strict_sub_structure_keys=True,
                                sub_structure={},
                            ),
                        },
                    ),
                    "postconditions": _validators,
                    "return": SemanticStructure(
                        type="property",
                    ),
                },
            ),
        },
    ),
    "ValueFunction": SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure={
                    "name": SemanticStructure(
                        sub_structure=SemanticStructure(
                            index_key_fn=lambda value: f"ValueFunction:{value}:def",
                            type="function",
                            modifier=[Modifier.definition],
                        ),
                    ),
                    "namespace": _namespace,
                },
            ),
            "spec": SemanticStructure(
                strict_sub_structure_keys=True,
                sub_structure={
                    "preconditions": _validators,
                    "locals": SemanticStructure(
                        type="property",
                        sub_structure=SemanticStructure(
                            sub_structure=SemanticStructure(
                                type="property",
                            ),
                        ),
                    ),
                    "return": SemanticStructure(
                        type="property",
                        sub_structure=SemanticStructure(
                            sub_structure=SemanticStructure(
                                type="property",
                            ),
                        ),
                    ),
                },
            ),
        },
    ),
    "Workflow": SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure=SemanticStructure(
                    sub_structure={
                        "name": SemanticStructure(
                            sub_structure=SemanticStructure(
                                index_key_fn=lambda value: f"Workflow:{value}:def",
                                type="class",
                                modifier=[Modifier.definition],
                            ),
                        ),
                        "namespace": _namespace,
                    },
                ),
            ),
            "spec": SemanticStructure(
                strict_sub_structure_keys=True,
                sub_structure={
                    "crdRef": SemanticStructure(
                        type="typeParameter",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "apiGroup": SemanticStructure(
                                type="parameter",
                                sub_structure=SemanticStructure(type="namespace"),
                            ),
                            "version": SemanticStructure(
                                type="parameter",
                                sub_structure=SemanticStructure(type="string"),
                            ),
                            "kind": SemanticStructure(
                                type="parameter",
                                sub_structure=SemanticStructure(type="class"),
                            ),
                        },
                    ),
                    "steps": SemanticStructure(
                        sub_structure=SemanticStructure(
                            sub_structure=SemanticStructure(
                                local_key_fn=step_path_indexer,
                                strict_sub_structure_keys=True,
                                sub_structure={
                                    "label": SemanticStructure(
                                        type="property",
                                        sub_structure=SemanticStructure(
                                            local_key_fn=lambda value: f"label:{value}",
                                            type="event",
                                            modifier=[Modifier.definition],
                                        ),
                                    ),
                                    "ref": _logic_ref,
                                    "refSwitch": SemanticStructure(
                                        type="property",
                                        strict_sub_structure_keys=True,
                                        sub_structure=SemanticStructure(
                                            type="property",
                                            strict_sub_structure_keys=True,
                                            sub_structure={
                                                "switchOn": SemanticStructure(
                                                    type="property"
                                                ),
                                                "cases": SemanticStructure(
                                                    type="property",
                                                    sub_structure=SemanticStructure(
                                                        strict_sub_structure_keys=True,
                                                        sub_structure=SemanticStructure(
                                                            strict_sub_structure_keys=True,
                                                            field_index_key_fn=function_ref_indexer,
                                                            sub_structure={
                                                                "case": SemanticStructure(
                                                                    type="property",
                                                                    sub_structure=SemanticStructure(
                                                                        type="string"
                                                                    ),
                                                                ),
                                                                "default": SemanticStructure(
                                                                    type="property",
                                                                    sub_structure=SemanticStructure(
                                                                        type="number"
                                                                    ),
                                                                ),
                                                                "kind": SemanticStructure(
                                                                    type="property",
                                                                    sub_structure=SemanticStructure(
                                                                        type="type"
                                                                    ),
                                                                ),
                                                                "name": SemanticStructure(
                                                                    type="property",
                                                                    sub_structure=SemanticStructure(
                                                                        type="function",
                                                                    ),
                                                                ),
                                                            },
                                                        ),
                                                    ),
                                                ),
                                            },
                                        ),
                                    ),
                                    "inputs": _function_inputs,
                                    "skipIf": SemanticStructure(
                                        type="property",
                                        sub_structure=SemanticStructure(
                                            type="string",
                                        ),
                                    ),
                                    "forEach": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "for_each",
                                        sub_structure={
                                            "itemIn": SemanticStructure(
                                                type="argument",
                                            ),
                                            "inputKey": SemanticStructure(
                                                type="argument",
                                                sub_structure=SemanticStructure(
                                                    type="parameter",
                                                    local_key_fn=lambda value: f"input:{value}",
                                                ),
                                            ),
                                            "state": SemanticStructure(type="property"),
                                            "condition": SemanticStructure(
                                                type="property",
                                                strict_sub_structure_keys=True,
                                                sub_structure={
                                                    "type": SemanticStructure(
                                                        type="property",
                                                        sub_structure=SemanticStructure(
                                                            type="type",
                                                        ),
                                                    ),
                                                    "name": SemanticStructure(
                                                        type="property",
                                                    ),
                                                },
                                            ),
                                        },
                                    ),
                                    "condition": SemanticStructure(
                                        type="property",
                                        strict_sub_structure_keys=True,
                                        sub_structure={
                                            "type": SemanticStructure(
                                                type="property",
                                                sub_structure=SemanticStructure(
                                                    type="type",
                                                ),
                                            ),
                                            "name": SemanticStructure(
                                                type="property",
                                            ),
                                        },
                                    ),
                                    "state": SemanticStructure(type="property"),
                                },
                            ),
                        ),
                    ),
                    "status": SemanticStructure(
                        type="property",
                        strict_sub_structure_keys=True,
                        sub_structure={
                            "conditions": SemanticStructure(
                                type="property",
                                sub_structure=SemanticStructure(
                                    strict_sub_structure_keys=True,
                                    sub_structure={
                                        "type": SemanticStructure(
                                            type="property",
                                            sub_structure=SemanticStructure(
                                                type="type"
                                            ),
                                        ),
                                        "name": SemanticStructure(type="property"),
                                        "step": SemanticStructure(type="property"),
                                    },
                                ),
                            ),
                            "state": SemanticStructure(type="property"),
                        },
                    ),
                },
            ),
        },
    ),
    "FunctionTest": SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure=SemanticStructure(
                    sub_structure={
                        "name": SemanticStructure(
                            sub_structure=SemanticStructure(
                                index_key_fn=lambda value: f"FunctionTest:{value}:def",
                                type="function",
                                modifier=[Modifier.definition],
                            ),
                        ),
                        "namespace": _namespace,
                    },
                ),
            ),
            "spec": SemanticStructure(
                local_key_fn=lambda value: "spec",
                strict_sub_structure_keys=True,
                sub_structure={
                    "functionRef": _logic_ref,
                    "currentResource": SemanticStructure(
                        type="property",
                        local_key_fn=lambda value: "current-resource",
                        sub_structure=SemanticStructure(
                            local_key_fn=lambda value: "current-value",
                        ),
                    ),
                    "inputs": _function_inputs,
                    "testCases": SemanticStructure(
                        type="property",
                        local_key_fn=lambda value: "test-cases",
                        sub_structure=SemanticStructure(
                            sub_structure=SemanticStructure(
                                type="property",
                                local_key_fn=lambda value: "test-case-body",
                                sub_structure={
                                    "label": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "label",
                                        sub_structure=SemanticStructure(type="string"),
                                    ),
                                    "variant": SemanticStructure(
                                        type="property",
                                        sub_structure=SemanticStructure(type="number"),
                                    ),
                                    "skip": SemanticStructure(
                                        type="property",
                                        sub_structure=SemanticStructure(type="number"),
                                    ),
                                    "currentResource": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "current-resource",
                                        sub_structure=SemanticStructure(
                                            local_key_fn=lambda value: "current-resource-value",
                                        ),
                                    ),
                                    "overlayResource": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "overlay-resource",
                                        sub_structure=SemanticStructure(
                                            local_key_fn=lambda value: "overlay-resource-value",
                                        ),
                                    ),
                                    "inputOverrides": _function_input_overrides,
                                    "expectResource": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "expected-resource",
                                        sub_structure=SemanticStructure(
                                            local_key_fn=lambda value: "expected-value",
                                        ),
                                    ),
                                    "expectOutcome": _validators,
                                    "expectReturn": SemanticStructure(
                                        type="property",
                                        local_key_fn=lambda value: "expected-return",
                                        sub_structure=SemanticStructure(
                                            local_key_fn=lambda value: "expected-value",
                                        ),
                                    ),
                                    "expectDelete": SemanticStructure(
                                        type="property",
                                        sub_structure=SemanticStructure(
                                            type="number",
                                        ),
                                    ),
                                },
                            ),
                        ),
                    ),
                    "expectResource": SemanticStructure(
                        type="property",
                        local_key_fn=lambda value: "expected_resource",
                        sub_structure=SemanticStructure(
                            local_key_fn=lambda value: "expected_value",
                        ),
                    ),
                    "expectOutcome": _validators,
                    "expectReturn": SemanticStructure(
                        type="property",
                        local_key_fn=lambda value: "expected_return",
                        sub_structure=SemanticStructure(
                            local_key_fn=lambda value: "expected_value",
                        ),
                    ),
                },
            ),
        },
    ),
    "ResourceTemplate": SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure=SemanticStructure(
                    sub_structure={
                        "name": SemanticStructure(
                            sub_structure=SemanticStructure(
                                index_key_fn=lambda value: f"ResourceTemplate:{value}:def",
                                type="function",
                                modifier=[Modifier.definition],
                            ),
                        ),
                        "namespace": _namespace,
                    },
                ),
            ),
            "spec": SemanticStructure(
                strict_sub_structure_keys=True,
                sub_structure={
                    "template": SemanticStructure(type="property"),
                },
            ),
        },
    ),
    ALL: SemanticStructure(
        sub_structure={
            "apiVersion": _api_version,
            "kind": _kind,
            "metadata": SemanticStructure(
                sub_structure=SemanticStructure(
                    sub_structure={
                        "name": SemanticStructure(
                            sub_structure=SemanticStructure(
                                modifier=[Modifier.definition],
                            ),
                        ),
                        "namespace": _namespace,
                    },
                ),
            ),
        },
    ),
}



from typing import Any, Literal, NamedTuple, Sequence
import asyncio

from lsprotocol import types

from celpy import celtypes

from koreo import cache
from koreo import registry
from koreo.result import (
    DepSkip,
    Ok,
    PermFail,
    Retry,
    Skip,
    UnwrappedOutcome,
    is_unwrapped_ok,
)


from koreo.function_test.run import run_function_test
from koreo.function_test.structure import FunctionTest

from . import constants


class CompareResult(NamedTuple):
    field: str
    expected: Any
    actual: Any


class FieldMismatchResult(NamedTuple):
    field: str
    severity: Literal["WARNING", "ERROR"]
    expected: bool
    actual: bool


class TestResults(NamedTuple):
    success: bool
    messages: list[str] | None = None
    resource_field_errors: list[CompareResult] | None = None
    outcome_fields_errors: list[CompareResult] | None = None
    input_mismatches: list[FieldMismatchResult] | None = None
    actual_resource: dict | None = None
    actual_return: dict | None = None
    missing_test_assertion: bool = False


async def run_function_tests(
    tests_to_run: set[str],
    functions_to_test: dict[type, Sequence[str]],
) -> tuple[dict[str, TestResults], list[types.LogMessageParams]]:
    function_tests = set[str]()
    for function_type, function_type_tests in functions_to_test.items():
        for function_name in function_type_tests:
            function_resource = registry.Resource(
                resource_type=function_type, name=function_name
            )
            subscribers = registry.get_subscribers(function_resource)
            for subscriber in subscribers:
                if subscriber.resource_type != FunctionTest.__qualname__:
                    continue

                function_tests.add(subscriber.name)

    all_tests = tests_to_run.union(function_tests)

    if not all_tests:
        return ({}, [])

    logs = []

    tasks = []
    try:
        async with asyncio.TaskGroup() as task_group:
            for test_key in all_tests:
                test = cache.get_resource_from_cache(
                    resource_class=FunctionTest, cache_key=test_key
                )
                if not test:
                    logs.append(
                        types.LogMessageParams(
                            type=types.MessageType.Error,
                            message=f"Failed to find FunctionTest ('{test_key}') in Koreo cache.",
                        )
                    )
                    continue

                tasks.append(
                    task_group.create_task(
                        _run_function_test(test_key, test), name=test_key
                    )
                )
    except:
        # exceptions handled for each task
        pass

    results = {task.get_name(): task.result() for task in tasks if task.done()}
    return (results, logs)


async def _run_function_test(
    test_name: str, test: UnwrappedOutcome[FunctionTest]
) -> TestResults:
    if not is_unwrapped_ok(test):
        return TestResults(
            success=False,
            messages=[f"{test.message}"],
            resource_field_errors=None,
            outcome_fields_errors=None,
            input_mismatches=None,
            actual_resource=None,
        )

    if not is_unwrapped_ok(test.function_under_test):
        return TestResults(
            success=False,
            messages=[f"{test.function_under_test}"],
            resource_field_errors=None,
            outcome_fields_errors=None,
            input_mismatches=None,
            actual_resource=None,
        )

    field_mismatches = _check_inputs(
        inputs=test.inputs,
        dynamic_input_keys=test.function_under_test.dynamic_input_keys,
    )

    success = True

    if field_mismatches:
        success = success and not any(
            mismatch.severity == "ERROR" for mismatch in field_mismatches
        )

    try:
        test_outcome = await run_function_test(location=test_name, function_test=test)
    except Exception as err:
        return TestResults(
            success=False,
            messages=[f"Unknown {err}."],
            resource_field_errors=None,
            outcome_fields_errors=None,
            input_mismatches=field_mismatches,
            actual_resource=None,
        )

    messages = []
    success = True

    if not test_outcome.test_results:
        messages.append("No tests run.")
    else:
        error_messages = []
        for result in test_outcome.test_results:
            if result.test_pass:
                continue

            if result.message:
                error_messages.append(f"{result.label}: {result.message}")
                continue

            if result.differences:
                error_messages.append(f"{result.label}: {result.differences}")

        if error_messages:
            messages.append(f"### Test Failures\n{'\n'.join(error_messages)}")
            success = False
        else:
            messages.append(f"### {len(test_outcome.test_results)} Tests passed")

        return TestResults(
            success=success,
            messages=messages,
            resource_field_errors=None,
            outcome_fields_errors=None,
            input_mismatches=field_mismatches,
            actual_resource=None,
        )

    return TestResults(
        success=success,
        messages=messages,
        resource_field_errors=None,
        outcome_fields_errors=None,
        input_mismatches=field_mismatches,
        actual_resource=None,
    )

    resource_field_errors = []
    outcome_fields_errors = []
    actual_return = None

    match test_outcome.outcome:
        case DepSkip(message=message, location=location):
            if not isinstance(test_outcome.expected_outcome, DepSkip):
                success = False
                messages.append(
                    f"Unexpected DepSkip(message='{message}', location='{location}'."
                )
        case Skip(message=message, location=location):
            if not isinstance(test_outcome.expected_outcome, Skip):
                success = False
                messages.append(
                    f"Unexpected Skip(message='{message}', location='{location}')."
                )
        case PermFail(message=message, location=location):
            if not isinstance(test_outcome.expected_outcome, PermFail):
                success = False
                messages.append(
                    f"Unexpected PermFail(message='{message}', location='{location}')."
                )
        case Retry(message=message, delay=delay, location=location):
            if test_outcome.expected_outcome is not None and not isinstance(
                test_outcome.expected_outcome, Retry
            ):
                messages.append(f"{test_outcome.expected_outcome}")
                success = False
                messages.append(
                    f"Unexpected Retry(message='{message}', delay={delay}, location='{location}')."
                )
            elif test_outcome.expected_return:
                success = False
                messages.append(
                    "Can not assert expected return-value when resource "
                    "modifications requested. Ensure currentResource matches "
                    "materializer."
                )

            resource_field_errors = _check_value(
                actual=test_outcome.actual_resource,
                expected=test_outcome.expected_resource,
            )
            if resource_field_errors:
                success = False

        case Ok(data=return_value) | return_value:
            if test_outcome.expected_outcome is not None and not isinstance(
                test_outcome.expected_outcome, Ok
            ):
                success = False
                messages.append("Unexpected Ok(...).")

            if test_outcome.expected_resource:
                success = False
                messages.append(
                    "Can not assert expectResource unless changes requested."
                )

            actual_return = return_value

            outcome_fields_errors = _check_value(
                actual=return_value, expected=test_outcome.expected_return
            )

            if outcome_fields_errors:
                success = False

    missing_test_assertion = False
    if not (
        test_outcome.expected_resource
        or test_outcome.expected_outcome
        or test_outcome.expected_return
    ):
        success = False
        missing_test_assertion = True
        messages.append("Must define expectResource, expectOutcome, or expectReturn.")

    return TestResults(
        success=success,
        messages=messages,
        resource_field_errors=resource_field_errors,
        outcome_fields_errors=outcome_fields_errors,
        input_mismatches=field_mismatches,
        actual_resource=test_outcome.actual_resource,
        actual_return=actual_return,
        missing_test_assertion=missing_test_assertion,
    )


def _check_inputs(
    inputs: celtypes.MapType | None,
    dynamic_input_keys: set[str],
) -> list[FieldMismatchResult]:
    provided_keys = set[str]()
    if inputs:
        provided_keys.update(f"inputs.{input_key}" for input_key in inputs.keys())

    first_tier_inputs = set(
        f"inputs.{match.group('name')}"
        for match in (
            constants.INPUT_NAME_PATTERN.match(key) for key in dynamic_input_keys
        )
        if match
    )

    mismatches = []
    for missing in first_tier_inputs.difference(provided_keys):
        mismatches.append(
            FieldMismatchResult(
                field=missing,
                severity="ERROR",
                actual=False,
                expected=True,
            )
        )

    for extra in provided_keys.difference(first_tier_inputs):
        mismatches.append(
            FieldMismatchResult(
                field=extra,
                severity="WARNING",
                actual=True,
                expected=False,
            )
        )

    return mismatches


def _check_value(actual: dict | None, expected: dict | None) -> list[CompareResult]:
    if expected is None:
        return []

    mismatches: list[CompareResult] = []

    if actual is None:
        mismatches.append(CompareResult(field="", actual="missing", expected="..."))
        return mismatches

    mismatches.extend(_dict_compare(base_field="", actual=actual, expected=expected))

    return mismatches


def _dict_compare(base_field: str, actual: dict, expected: dict) -> list[CompareResult]:
    differences: list[CompareResult] = []

    if base_field:
        base_prefix = f"{base_field}."
    else:
        base_prefix = ""

    actual_keys = set(actual.keys())
    expected_keys = set(expected.keys())

    for key in actual_keys.difference(expected_keys):
        differences.append(
            CompareResult(
                field=f"{base_prefix}{key}",
                actual="...",
                expected="missing",
            )
        )

    for key in expected_keys.difference(actual_keys):
        differences.append(
            CompareResult(
                field=f"{base_prefix}{key}",
                actual="missing",
                expected="...",
            )
        )

    for mutual_key in actual_keys.intersection(expected_keys):
        actual_value = actual[mutual_key]
        expected_value = expected[mutual_key]

        key = f"{base_prefix}{mutual_key}"

        if type(actual_value) != type(expected_value):
            differences.append(
                CompareResult(
                    field=key,
                    actual=f"{type(actual_value)}",
                    expected=f"{type(expected_value)}",
                )
            )
            continue

        if isinstance(actual_value, dict):
            differences.extend(
                _dict_compare(
                    base_field=key, actual=actual_value, expected=expected_value
                )
            )
            continue

        differences.extend(
            _values_match(field=key, actual=actual_value, expected=expected_value)
        )

    return differences


def _values_match(field: str, actual, expected) -> list[CompareResult]:
    if type(actual) != type(expected):
        return [
            CompareResult(
                field=field,
                actual=f"{type(actual)}",
                expected=f"{type(expected)}",
            )
        ]

    if isinstance(actual, dict):
        return _dict_compare(base_field=field, actual=actual, expected=expected)

    if isinstance(actual, list):
        if len(actual) != len(expected):
            return [
                CompareResult(
                    field=field,
                    actual=f"{len(actual)} items",
                    expected=f"{len(expected)} items",
                )
            ]

        mismatches = []
        for idx, (actual_value, expected_value) in enumerate(zip(actual, expected)):
            mismatches.extend(
                _values_match(
                    field=f"{field}[{idx}]",
                    actual=actual_value,
                    expected=expected_value,
                )
            )

        return mismatches

    if actual == expected:
        return []

    return [
        CompareResult(
            field=field,
            actual=f"{actual}",
            expected=f"{expected}",
        )
    ]



from typing import NamedTuple
import re

from .semantics import (
    Modifier,
    NodeDiagnostic,
    SemanticNode,
    Position,
    Severity,
    TokenType,
)

SYMBOL = re.compile(r"[\w]+")
QUOTED = re.compile(r"(?P<quote>['\"])(?P<string>.*?)(?P=quote)")
OP = re.compile(r"->|[\{\}\(\)\.,+:*-=\[\]^$!<>|?&\\]")
SPACE = re.compile(r"\s+")

NUMBER = re.compile(r"\d+")

KEYWORDS = {"has", "all", "exists", "exists_one", "map", "filter"}


class Token(NamedTuple):
    position: Position
    start_rel: Position

    text: str

    token_type: TokenType
    token_modifiers: list[Modifier]


def is_dquote(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == '"' and token.token_type == "operator"


def is_squote(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == "'" and token.token_type == "operator"


def is_lparen(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == "(" and token.token_type == "operator"


def is_lbrace(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == "{" and token.token_type == "operator"


def is_rbrace(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == "}" and token.token_type == "operator"


def is_colon(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == ":" and token.token_type == "operator"


def is_comma(token: Token | None) -> bool:
    if token is None:
        return False

    return token.text == "," and token.token_type == "operator"


def parse(
    cel_expression: list[str],
    anchor_base_pos: Position,
    seed_line: int = 0,
    seed_offset: int = 0,
    abs_offset: int = 0,
) -> list[SemanticNode]:
    """Convert the given expression into a list of tokens"""
    tokens = lex(
        cel_expression=cel_expression,
        seed_line=seed_line,
        seed_offset=seed_offset,
        abs_offset=abs_offset,
    )

    return _extract_semantic_structure(tokens, anchor_base_pos)


def _extract_semantic_structure(
    tokens: list[Token], anchor_base_pos: Position
) -> list[SemanticNode]:
    """Given a list of tokens, determine their type and modifiers."""

    def next(idx):
        """Get the next token, if possible"""
        if idx >= len(tokens) - 1:
            return None

        return tokens[idx + 1]

    in_brace = False
    in_dquote = False
    in_squote = False

    nodes: list[SemanticNode] = []
    for idx, token in enumerate(tokens):
        if token.token_type == "operator":
            node_diagnostic = None
            if is_lbrace(token):
                in_brace = True

            elif is_rbrace(token):
                in_brace = False

            elif is_dquote(token):
                in_dquote = not in_dquote

            elif is_squote(token):
                in_squote = not in_squote

            elif is_comma(token) and is_rbrace(next(idx)):
                node_diagnostic = NodeDiagnostic(
                    message="Trailing commas are unsupported.", severity=Severity.error
                )

            nodes.append(
                SemanticNode(
                    position=token.position,
                    anchor_rel=Position(
                        line=token.start_rel.line + anchor_base_pos.line,
                        character=token.start_rel.character,
                    ),
                    length=len(token.text),
                    node_type=token.token_type,
                    modifier=token.token_modifiers,
                    diagnostic=node_diagnostic,
                )
            )

            continue

        token_type = ""
        if in_dquote or in_squote:
            if is_colon(next(idx + 1)) and in_brace:
                token_type = "property"
            else:
                token_type = "string"

        elif token.text in KEYWORDS:
            token_type = "keyword"

        elif is_lparen(next(idx)):
            token_type = "function"

        else:
            if NUMBER.match(token.text):
                token_type = "number"
            else:
                token_type = "variable"

        nodes.append(
            SemanticNode(
                position=token.position,
                anchor_rel=Position(
                    line=token.start_rel.line + anchor_base_pos.line,
                    character=token.start_rel.character,
                ),
                length=len(token.text),
                node_type=token_type,
                modifier=token.token_modifiers,
            )
        )

    return nodes


def lex(
    cel_expression: list[str],
    seed_line: int = 0,
    seed_offset: int = 0,
    abs_offset: int = 0,
) -> list[Token]:
    """Convert the given document into a list of tokens"""
    tokens = []

    prev_line = 0

    for current_line, line in enumerate(cel_expression, start=seed_line):
        prev_offset = 0
        current_offset = seed_offset
        chars_left = len(line)

        if not line.strip():
            seed_offset = 0
            continue

        while line:
            if (match := SPACE.match(line)) is not None:
                # Skip whitespace
                current_offset += len(match.group(0))
                line = line[match.end() :]

            elif (match := QUOTED.match(line)) is not None:
                quote_group_len = len(match.group("quote"))
                string_len = len(match.group("string"))

                tokens.append(
                    Token(
                        position=Position(
                            line=current_line - prev_line,
                            character=current_offset - prev_offset,
                        ),
                        start_rel=Position(
                            line=current_line,
                            character=abs_offset + current_offset,
                        ),
                        text=match.group("quote"),
                        token_type="operator",
                        token_modifiers=[],
                    ),
                )

                if string_len:
                    tokens.append(
                        Token(
                            position=Position(line=0, character=quote_group_len),
                            start_rel=Position(
                                line=current_line,
                                character=abs_offset + current_offset + quote_group_len,
                            ),
                            text=match.group("string"),
                            token_type="string",
                            token_modifiers=[],
                        )
                    )

                tokens.append(
                    Token(
                        position=Position(
                            line=0,
                            character=string_len if string_len else quote_group_len,
                        ),
                        start_rel=Position(
                            line=current_line,
                            character=abs_offset
                            + current_offset
                            + quote_group_len
                            + string_len,
                        ),
                        text=match.group("quote"),
                        token_type="operator",
                        token_modifiers=[],
                    ),
                )

                line = line[match.end() :]
                prev_line = current_line
                # First quote + quoted string
                current_offset += quote_group_len + string_len
                prev_offset = current_offset
                # Closing quote
                current_offset += quote_group_len

            elif (match := SYMBOL.match(line)) is not None:
                tokens.append(
                    Token(
                        position=Position(
                            line=current_line - prev_line,
                            character=current_offset - prev_offset,
                        ),
                        start_rel=Position(
                            line=current_line,
                            character=abs_offset + current_offset,
                        ),
                        text=match.group(0),
                        token_type="",
                        token_modifiers=[],
                    )
                )

                line = line[match.end() :]
                prev_offset = current_offset
                prev_line = current_line
                current_offset += len(match.group(0))

            elif (match := OP.match(line)) is not None:
                tokens.append(
                    Token(
                        position=Position(
                            line=current_line - prev_line,
                            character=current_offset - prev_offset,
                        ),
                        start_rel=Position(
                            line=current_line,
                            character=abs_offset + current_offset,
                        ),
                        text=match.group(0),
                        token_type="operator",
                        token_modifiers=[],
                    )
                )

                line = line[match.end() :]
                prev_offset = current_offset
                prev_line = current_line
                current_offset += len(match.group(0))

            else:
                raise RuntimeError(f"No match: {line!r}")

            # Make sure we don't hit an infinite loop
            if (n := len(line)) == chars_left:
                raise RuntimeError("Inifite loop detected")
            else:
                chars_left = n

        prev_line = current_line
        seed_offset = 0
        abs_offset = 0

    return tokens



from functools import reduce
from typing import Any, Generator, NamedTuple, Sequence
import operator

import yaml.parser
import yaml.scanner
import yaml.loader

from pygls.workspace import TextDocument
from lsprotocol import types

from koreo import cache

from koreo_tooling.indexing import IndexingLoader, STRUCTURE_KEY
from koreo_tooling.indexing.semantics import (
    SemanticAnchor,
    SemanticNode,
    TokenTypes,
    compute_abs_range,
    extract_diagnostics,
    flatten,
    generate_key_range_index,
)

from koreo_tooling import constants
from koreo_tooling.langserver.rangers import block_range_extract

TypeIndex = {key: idx for idx, key in enumerate(TokenTypes)}


class SemanticRangeIndex(NamedTuple):
    uri: str
    name: str
    range: types.Range
    version: str | None = None


class ProccessResults(NamedTuple):
    semantic_tokens: Sequence[int] | None = None
    semantic_range_index: Sequence[SemanticRangeIndex] | None = None
    logs: Sequence[types.LogMessageParams] | None = None
    diagnostics: Sequence[types.Diagnostic] | None = None


async def process_file(doc: TextDocument) -> ProccessResults:
    uri = doc.uri

    semantic_tokens: list[int] = []
    semantic_range_index: list[SemanticRangeIndex] = []
    logs: list[types.LogMessageParams] = []
    diagnostics: list[types.Diagnostic] = []

    yaml_blocks = _load_all_yamls(doc.source, Loader=IndexingLoader, doc=doc)
    for yaml_block in yaml_blocks:
        match yaml_block:
            case YamlParseError(
                err=err, problem_pos=problem_pos, context_pos=context_pos
            ):
                logs.append(
                    types.LogMessageParams(
                        type=types.MessageType.Error,
                        message=f"Failed to parse YAML around {problem_pos} / {context_pos}",
                    )
                )

                last_line = len(doc.lines) - 1

                if problem_pos:
                    diagnostics.append(
                        types.Diagnostic(
                            message=err,
                            severity=types.DiagnosticSeverity.Error,
                            range=types.Range(
                                start=types.Position(
                                    line=problem_pos.line, character=0
                                ),
                                end=types.Position(
                                    line=problem_pos.line,
                                    character=len(
                                        doc.lines[min(problem_pos.line, last_line)]
                                    ),
                                ),
                            ),
                        )
                    )

                if context_pos:
                    diagnostics.append(
                        types.Diagnostic(
                            message=err,
                            severity=types.DiagnosticSeverity.Error,
                            range=types.Range(
                                start=types.Position(
                                    line=context_pos.line, character=0
                                ),
                                end=types.Position(
                                    line=context_pos.line,
                                    character=len(
                                        doc.lines[min(context_pos.line, last_line)]
                                    ),
                                ),
                            ),
                        )
                    )

                break

            case (block_hash, block_doc):
                block_result = await _process_block(
                    uri=uri,
                    yaml_block=block_doc,
                    doc=doc,
                    block_hash=block_hash,
                )

                if block_result.logs:
                    logs.extend(block_result.logs)

                if block_result.semantic_tokens:
                    semantic_tokens.extend(
                        value
                        for token in block_result.semantic_tokens
                        for value in token
                    )

                if block_result.semantic_range_index:
                    semantic_range_index.extend(block_result.semantic_range_index)

                    # TODO: Extract def and tag version in index?

                if block_result.diagnostics:
                    diagnostics.extend(block_result.diagnostics)

    return ProccessResults(
        semantic_range_index=semantic_range_index,
        semantic_tokens=semantic_tokens,
        logs=logs,
        diagnostics=diagnostics,
    )


class BlockResults(NamedTuple):
    semantic_tokens: Generator[tuple[int], None, None] | None = None
    semantic_range_index: Generator[SemanticRangeIndex, None, None] | None = None
    logs: Sequence[types.LogMessageParams] | None = None
    diagnostics: Sequence[types.Diagnostic] | None = None


async def _process_block(
    uri: str, yaml_block: dict, doc: TextDocument, block_hash: str
) -> BlockResults:
    try:
        api_version = yaml_block.get("apiVersion")
        kind = yaml_block.get("kind")
    except:
        return BlockResults(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Info,
                    message="Skipping empty block",
                )
            ]
        )

    semantic_anchor: SemanticAnchor | None = yaml_block.get(STRUCTURE_KEY)
    if not semantic_anchor:
        block_result = BlockResults(
            logs=[
                types.LogMessageParams(
                    type=types.MessageType.Info,
                    message="Failed to load semantic anchor",
                )
            ]
        )
        return block_result

    semantic_range_index = (
        SemanticRangeIndex(uri=uri, name=node_key, range=node_range, version=block_hash)
        for node_key, node_range in generate_key_range_index(semantic_anchor)
    )

    flattened = flatten(semantic_anchor)
    semantic_tokens = _to_lsp_semantics(flattened)

    logs: list[types.LogMessageParams] = []
    diagnostics: list[types.Diagnostic] = []

    for node in extract_diagnostics(flattened):
        diagnostics.append(
            types.Diagnostic(
                message=node.diagnostic.message,
                severity=types.DiagnosticSeverity.Error,  # TODO: Map internal to LSP
                range=compute_abs_range(node, semantic_anchor),
            )
        )

    if api_version not in {
        constants.API_VERSION,
        constants.CRD_API_VERSION,
    }:
        api_version_block = block_range_extract(
            search_key="api_version",
            search_nodes=semantic_anchor.children,
            anchor=semantic_anchor,
        )
        match api_version_block:
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case None:
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case _:
                resource_range = compute_abs_range(api_version_block, semantic_anchor)

        diagnostics.append(
            types.Diagnostic(
                message=f"'{kind}.{api_version}' is not a Koreo Resource.",
                severity=types.DiagnosticSeverity.Information,
                range=resource_range,
            )
        )

        block_result = BlockResults(
            semantic_range_index=semantic_range_index,
            semantic_tokens=semantic_tokens,
            diagnostics=diagnostics,
        )
        return block_result

    if kind not in constants.PREPARE_MAP and kind != constants.CRD_KIND:
        kind_version_block = block_range_extract(
            search_key="kind",
            search_nodes=semantic_anchor.children,
            anchor=semantic_anchor,
        )
        match kind_version_block:
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case None:
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case _:
                resource_range = compute_abs_range(kind_version_block, semantic_anchor)

        diagnostics.append(
            types.Diagnostic(
                message=f"'{kind}' is not a supported Koreo Resource.",
                severity=types.DiagnosticSeverity.Information,
                range=resource_range,
            )
        )
        block_result = BlockResults(
            semantic_range_index=semantic_range_index,
            semantic_tokens=semantic_tokens,
            diagnostics=diagnostics,
        )
        return block_result

    resource_class, preparer = constants.PREPARE_MAP[kind]
    metadata = yaml_block.get("metadata", {})
    name = metadata.get("name", "missing-name")

    resource_version = f"{block_hash}"
    metadata["resourceVersion"] = resource_version

    raw_spec = yaml_block.get("spec", {})

    cached_system_data = {
        "uri": uri,
        "anchor": semantic_anchor,
    }

    try:
        await cache.prepare_and_cache(
            resource_class=resource_class,
            preparer=preparer,
            metadata=metadata,
            spec=raw_spec,
            _system_data=cached_system_data,
        )

    except Exception as err:
        resource_name_label = block_range_extract(
            search_key=f"{kind}:{name}:def",
            search_nodes=semantic_anchor.children,
            anchor=semantic_anchor,
        )
        match resource_name_label:
            case list(block_diagnostics):
                diagnostics.extend(block_diagnostics)
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case None:
                resource_range = _block_range(semantic_anchor=semantic_anchor, doc=doc)
            case _:
                resource_range = compute_abs_range(resource_name_label, semantic_anchor)

        diagnostics.append(
            types.Diagnostic(
                message=f"FAILED TO Extract ('{kind}.{api_version}') from {name} ({err}).",
                severity=types.DiagnosticSeverity.Error,
                range=resource_range,
            )
        )

    block_result = BlockResults(
        semantic_range_index=semantic_range_index,
        semantic_tokens=semantic_tokens,
        diagnostics=diagnostics,
        logs=logs,
    )
    return block_result


class YamlParseError(NamedTuple):
    err: str
    problem_pos: types.Position | None
    context_pos: types.Position | None


def _load_all_yamls(
    stream, Loader, doc
) -> Generator[tuple[str, dict] | YamlParseError, Any, None]:
    """
    Parse all YAML documents in a stream
    and produce corresponding Python objects.
    """
    loader = Loader(stream, doc=doc)
    try:
        while loader.check_data():
            try:
                yield loader.get_data()
            except (yaml.scanner.ScannerError, yaml.parser.ParserError) as err:
                problem_pos = None
                if err.problem_mark:
                    problem_pos = types.Position(
                        line=err.problem_mark.line, character=err.problem_mark.column
                    )

                context_pos = None
                if err.context_mark:
                    context_pos = types.Position(
                        line=err.context_mark.line, character=err.context_mark.column
                    )

                yield YamlParseError(
                    err=f"{err}", problem_pos=problem_pos, context_pos=context_pos
                )

    finally:
        loader.dispose()


def _to_lsp_semantics(nodes: Sequence[SemanticNode]) -> Generator[tuple, None, None]:
    for node in nodes:
        yield (
            node.position.line,
            node.position.character,
            node.length,
            TypeIndex[node.node_type],
            reduce(operator.or_, node.modifier, 0) if node.modifier else 0,
        )


def _block_range(semantic_anchor: SemanticAnchor, doc: TextDocument) -> types.Range:
    return types.Range(
        start=types.Position(
            line=semantic_anchor.abs_position.line,
            character=semantic_anchor.abs_position.character,
        ),
        end=types.Position(
            line=semantic_anchor.abs_position.line,
            character=len(doc.lines[semantic_anchor.abs_position.line]),
        ),
    )



import re

from koreo.function_test.prepare import prepare_function_test
from koreo.function_test.structure import FunctionTest
from koreo.resource_function.prepare import prepare_resource_function
from koreo.resource_function.structure import ResourceFunction
from koreo.resource_template.prepare import prepare_resource_template
from koreo.resource_template.structure import ResourceTemplate
from koreo.value_function.prepare import prepare_value_function
from koreo.value_function.structure import ValueFunction
from koreo.workflow.prepare import prepare_workflow
from koreo.workflow.structure import Workflow


API_VERSION = "koreo.dev/v1beta1"

PREPARE_MAP = {
    "FunctionTest": (FunctionTest, prepare_function_test),
    "ResourceFunction": (ResourceFunction, prepare_resource_function),
    "ResourceTemplate": (ResourceTemplate, prepare_resource_template),
    "ValueFunction": (ValueFunction, prepare_value_function),
    "Workflow": (Workflow, prepare_workflow),
}

CRD_API_VERSION = "apiextensions.k8s.io/v1"
CRD_KIND = "CustomResourceDefinition"


RESOURCE_DEF = re.compile("(?P<kind>[A-Z][a-zA-Z0-9.]*):(?P<name>.*):def")

TOP_LEVEL_RESOURCE = re.compile("(?P<kind>[A-Z][a-zA-Z0-9.]*):(?P<name>.*)?:[dr]ef")

FUNCTION_TEST_NAME = re.compile("FunctionTest:(?P<name>.*)?:def")
RESOURCE_FUNCTION_NAME = re.compile("ResourceFunction:(?P<name>.*)?:def")
VALUE_FUNCTION_NAME = re.compile("ValueFunction:(?P<name>.*)?:def")
WORKFLOW_NAME = re.compile("Workflow:(?P<name>[^:]*)?:def")

FUNCTION_TEST_ANCHOR = re.compile("FunctionTest:(?P<name>.*)")
RESOURCE_FUNCTION_ANCHOR = re.compile("ResourceFunction:(?P<name>.*)")
VALUE_FUNCTION_ANCHOR = re.compile("ValueFunction:(?P<name>.*)")
WORKFLOW_ANCHOR = re.compile("Workflow:(?P<name>[^:]*)")

INPUT_NAME_PATTERN = re.compile("inputs.(?P<name>[^.]+).?")
PARENT_ROOT_PATTERN = re.compile("(?P<root>[^.]+).?")
